@@ -3,72 +3,117 @@ use strict;
 use warnings;
 use Carp ();
 use Scalar::Util ();
+use Module::Runtime ();
 
 BEGIN {
+  # use M::R to work around the 5.8 require bugs
   if ($] < 5.009_005) {
-    require MRO::Compat;
+    Module::Runtime::require_module('MRO::Compat');
   }
   else {
     require mro;
   }
 }
 
-our $VERSION = '0.10006';
+our $VERSION = '0.10007';
 $VERSION = eval $VERSION if $VERSION =~ /_/; # numify for warning-free dev releases
 
 # when changing minimum version don't forget to adjust Makefile.PL as well
 our $__minimum_xsa_version;
-BEGIN { $__minimum_xsa_version = '1.13' }
+BEGIN { $__minimum_xsa_version = '1.15' }
 
 our $USE_XS;
 # the unless defined is here so that we can override the value
 # before require/use, *regardless* of the state of $ENV{CAG_USE_XS}
 $USE_XS = $ENV{CAG_USE_XS}
-    unless defined $USE_XS;
+  unless defined $USE_XS;
+
+BEGIN {
+  package # hide from PAUSE
+    __CAG_ENV__;
+
+  die "Huh?! No minimum C::XSA version?!\n"
+    unless $__minimum_xsa_version;
+
+  local $@;
+  require constant;
+
+  # individual (one const at a time) imports so we are 5.6.2 compatible
+  # if we can - why not ;)
+  constant->import( NO_SUBNAME => eval {
+    Module::Runtime::require_module('Sub::Name')
+  } ? 0 : "$@" );
+
+  my $found_cxsa;
+  constant->import( NO_CXSA => ( !NO_SUBNAME() and eval {
+    Module::Runtime::require_module('Class::XSAccessor');
+    $found_cxsa = Class::XSAccessor->VERSION;
+    Class::XSAccessor->VERSION($__minimum_xsa_version);
+  } ) ? 0 : "$@" );
+
+  if (NO_CXSA() and $found_cxsa and !$ENV{CAG_OLD_XS_NOWARN}) {
+    warn(
+      'The installed version of Class::XSAccessor is too old '
+    . "(v$found_cxsa < v$__minimum_xsa_version). Please upgrade "
+    . "to instantly quadruple the performance of 'simple' accessors. "
+    . 'Set $ENV{CAG_OLD_XS_NOWARN} if you wish to disable this '
+    . "warning.\n"
+    );
+  }
+
+  constant->import( BROKEN_GOTO => ($] < '5.008009') ? 1 : 0 );
+
+  constant->import( UNSTABLE_DOLLARAT => ($] < '5.013002') ? 1 : 0 );
+
+  constant->import( TRACK_UNDEFER_FAIL => (
+    $INC{'Test/Builder.pm'} || $INC{'Test/Builder2.pm'}
+      and
+    $0 =~ m|^ x?t / .+ \.t $|x
+  ) ? 1 : 0 );
+}
 
 # Yes this method is undocumented
 # Yes it should be a private coderef like all the rest at the end of this file
 # No we can't do that (yet) because the DBIC-CDBI compat layer overrides it
 # %$*@!?&!&#*$!!!
 sub _mk_group_accessors {
-    my($self, $maker, $group, @fields) = @_;
-    my $class = Scalar::Util::blessed $self || $self;
+  my($self, $maker, $group, @fields) = @_;
+  my $class = length (ref ($self) ) ? ref ($self) : $self;
 
-    no strict 'refs';
-    no warnings 'redefine';
+  no strict 'refs';
+  no warnings 'redefine';
 
-    # So we don't have to do lots of lookups inside the loop.
-    $maker = $self->can($maker) unless ref $maker;
+  # So we don't have to do lots of lookups inside the loop.
+  $maker = $self->can($maker) unless ref $maker;
 
-    foreach (@fields) {
-        if( $_ eq 'DESTROY' ) {
-            Carp::carp("Having a data accessor named DESTROY in ".
-                       "'$class' is unwise.");
-        }
+  for (@fields) {
 
-        my ($name, $field) = (ref $_)
-            ? (@$_)
-            : ($_, $_)
-        ;
+    my ($name, $field) = (ref $_) ? (@$_) : ($_, $_);
 
-        my $alias = "_${name}_accessor";
+    Carp::croak("Illegal accessor name '$name'")
+      unless $name =~ /\A[A-Z_a-z][0-9A-Z_a-z]*\z/;
 
-        for my $meth ($name, $alias) {
+    Carp::carp("Having a data accessor named '$name' in '$class' is unwise.")
+      if $name =~ /\A(?: DESTROY | AUTOLOAD | CLONE )\z/x;
 
-            # the maker may elect to not return anything, meaning it already
-            # installed the coderef for us (e.g. lack of Sub::Name)
-            my $cref = $self->$maker($group, $field, $meth)
-                or next;
+    my $alias = "_${name}_accessor";
 
-            my $fq_meth = "${class}::${meth}";
+    for ($name, $alias) {
 
-            *$fq_meth = Sub::Name::subname($fq_meth, $cref);
-                #unless defined &{$class."\:\:$field"}
-        }
+      # the maker may elect to not return anything, meaning it already
+      # installed the coderef for us (e.g. lack of Sub::Name)
+      my $cref = $self->$maker($group, $field, $_)
+        or next;
+
+      my $fq_meth = "${class}::$_";
+
+      *$fq_meth = Sub::Name::subname($fq_meth, $cref);
+        #unless defined &{$class."\:\:$field"}
     }
+  }
 };
 
-# coderef is setup at the end for clarity
+# $gen_accessor coderef is setup at the end for clarity
 my $gen_accessor;
 
 =head1 NAME
@@ -120,10 +165,10 @@ be of the form [ $accessor, $field ].
 =cut
 
 sub mk_group_accessors {
-    my ($self, $group, @fields) = @_;
+  my ($self, $group, @fields) = @_;
 
-    $self->_mk_group_accessors('make_group_accessor', $group, @fields);
-    return;
+  $self->_mk_group_accessors('make_group_accessor', $group, @fields);
+  return;
 }
 
 =head2 mk_group_ro_accessors
@@ -145,9 +190,10 @@ rather than setting the value.
 =cut
 
 sub mk_group_ro_accessors {
-    my($self, $group, @fields) = @_;
+  my($self, $group, @fields) = @_;
 
-    $self->_mk_group_accessors('make_group_ro_accessor', $group, @fields);
+  $self->_mk_group_accessors('make_group_ro_accessor', $group, @fields);
+  return;
 }
 
 =head2 mk_group_wo_accessors
@@ -169,9 +215,10 @@ value rather than getting the value.
 =cut
 
 sub mk_group_wo_accessors {
-    my($self, $group, @fields) = @_;
+  my($self, $group, @fields) = @_;
 
-    $self->_mk_group_accessors('make_group_wo_accessor', $group, @fields);
+  $self->_mk_group_accessors('make_group_wo_accessor', $group, @fields);
+  return;
 }
 
 =head2 get_simple
@@ -190,7 +237,7 @@ name passed as an argument.
 =cut
 
 sub get_simple {
-    return $_[0]->{$_[1]};
+  $_[0]->{$_[1]};
 }
 
 =head2 set_simple
@@ -209,7 +256,7 @@ for the field name passed as an argument.
 =cut
 
 sub set_simple {
-    return $_[0]->{$_[1]} = $_[2];
+  $_[0]->{$_[1]} = $_[2];
 }
 
 
@@ -232,38 +279,30 @@ instances.
 =cut
 
 sub get_inherited {
-    my $class;
-
-    if ( defined( $class = Scalar::Util::blessed $_[0] ) ) {
-        if (Scalar::Util::reftype $_[0] eq 'HASH') {
-          return $_[0]->{$_[1]} if exists $_[0]->{$_[1]};
-        }
-        else {
-          Carp::croak('Cannot get inherited value on an object instance that is not hash-based');
-        }
+  if ( length (ref ($_[0]) ) ) {
+    if (Scalar::Util::reftype $_[0] eq 'HASH') {
+      return $_[0]->{$_[1]} if exists $_[0]->{$_[1]};
+      # everything in @_ is aliased, an assignment won't work
+      splice @_, 0, 1, ref($_[0]);
     }
     else {
-        $class = $_[0];
+      Carp::croak('Cannot get inherited value on an object instance that is not hash-based');
     }
+  }
 
-    no strict 'refs';
-    no warnings 'uninitialized';
+  # if we got this far there is nothing in the instance
+  # OR this is a class call
+  # in any case $_[0] contains the class name (see splice above)
+  no strict 'refs';
+  no warnings 'uninitialized';
 
-    my $cag_slot = '::__cag_'. $_[1];
-    return ${$class.$cag_slot} if defined(${$class.$cag_slot});
+  my $cag_slot = '::__cag_'. $_[1];
+  return ${$_[0].$cag_slot} if defined(${$_[0].$cag_slot});
 
-    # we need to be smarter about recalculation, as @ISA (thus supers) can very well change in-flight
-    my $cur_gen = mro::get_pkg_gen ($class);
-    if ( $cur_gen != ${$class.'::__cag_pkg_gen__'} ) {
-        @{$class.'::__cag_supers__'} = $_[0]->get_super_paths;
-        ${$class.'::__cag_pkg_gen__'} = $cur_gen;
-    }
+  do { return ${$_.$cag_slot} if defined(${$_.$cag_slot}) }
+    for $_[0]->get_super_paths;
 
-    for (@{$class.'::__cag_supers__'}) {
-        return ${$_.$cag_slot} if defined(${$_.$cag_slot});
-    };
-
-    return undef;
+  return undef;
 }
 
 =head2 set_inherited
@@ -287,17 +326,16 @@ hash-based object.
 =cut
 
 sub set_inherited {
-    if (defined Scalar::Util::blessed $_[0]) {
-        if (Scalar::Util::reftype $_[0] eq 'HASH') {
-            return $_[0]->{$_[1]} = $_[2];
-        } else {
-            Carp::croak('Cannot set inherited value on an object instance that is not hash-based');
-        };
+  if (length (ref ($_[0]) ) ) {
+    if (Scalar::Util::reftype $_[0] eq 'HASH') {
+      return $_[0]->{$_[1]} = $_[2];
     } else {
-        no strict 'refs';
-
-        return ${$_[0].'::__cag_'.$_[1]} = $_[2];
+      Carp::croak('Cannot set inherited value on an object instance that is not hash-based');
     };
+  }
+
+  no strict 'refs';
+  ${$_[0].'::__cag_'.$_[1]} = $_[2];
 }
 
 =head2 get_component_class
@@ -312,17 +350,17 @@ Returns: $value
 
 Gets the value of the specified component class.
 
-    __PACKAGE__->mk_group_accessors('component_class' => 'result_class');
+ __PACKAGE__->mk_group_accessors('component_class' => 'result_class');
 
-    $self->result_class->method();
+ $self->result_class->method();
 
-    ## same as
-    $self->get_component_class('result_class')->method();
+ ## same as
+ $self->get_component_class('result_class')->method();
 
 =cut
 
 sub get_component_class {
-    return $_[0]->get_inherited($_[1]);
+  $_[0]->get_inherited($_[1]);
 };
 
 =head2 set_component_class
@@ -338,25 +376,36 @@ Returns: $new_value
 Inherited accessor that automatically loads the specified class before setting
 it. This method will die if the specified class could not be loaded.
 
-    __PACKAGE__->mk_group_accessors('component_class' => 'result_class');
-    __PACKAGE__->result_class('MyClass');
+ __PACKAGE__->mk_group_accessors('component_class' => 'result_class');
+ __PACKAGE__->result_class('MyClass');
 
-    $self->result_class->method();
+ $self->result_class->method();
 
 =cut
 
 sub set_component_class {
-    if ($_[2]) {
-        local $^W = 0;
-        require Class::Inspector;
-        if (Class::Inspector->installed($_[2]) && !Class::Inspector->loaded($_[2])) {
-            eval "require $_[2]";
+  if (defined $_[2] and length $_[2]) {
+    # disable warnings, and prevent $_ being eaten away by a behind-the-scenes
+    # module loading
+    local ($^W, $_);
 
-            Carp::croak("Could not load $_[1] '$_[2]': ", $@) if $@;
-        };
-    };
+    if (__CAG_ENV__::UNSTABLE_DOLLARAT) {
+      my $err;
+      {
+        local $@;
+        eval { Module::Runtime::use_package_optimistically($_[2]) }
+          or $err = $@;
+      }
+      Carp::croak("Could not load $_[1] '$_[2]': $err") if defined $err;
 
-    return $_[0]->set_inherited($_[1], $_[2]);
+    }
+    else {
+      eval { Module::Runtime::use_package_optimistically($_[2]) }
+        or Carp::croak("Could not load $_[1] '$_[2]': $@");
+    }
+  };
+
+  $_[0]->set_inherited($_[1], $_[2]);
 };
 
 =head1 INTERNAL METHODS
@@ -372,7 +421,10 @@ inherited from. This is what drives the traversal done by L</get_inherited>.
 =cut
 
 sub get_super_paths {
-    return @{mro::get_linear_isa( ref($_[0]) || $_[0] )};
+  # get_linear_isa returns the class itself as the 1st element
+  # use @_ as a pre-allocated scratch array
+  (undef, @_) = @{mro::get_linear_isa( length( ref($_[0]) ) ? ref($_[0]) : $_[0] )};
+  @_;
 };
 
 =head2 make_group_accessor
@@ -526,62 +578,6 @@ it under the same terms as perl itself.
 ########################################################################
 ########################################################################
 
-BEGIN {
-
-  die "Huh?! No minimum C::XSA version?!\n"
-    unless $__minimum_xsa_version;
-
-  local $@;
-  my $err;
-
-
-  $err = eval { require Sub::Name; 1; } ? undef : do {
-    delete $INC{'Sub/Name.pm'};   # because older perls suck
-    $@;
-  };
-  *__CAG_ENV__::NO_SUBNAME = $err
-    ? sub () { $err }
-    : sub () { 0 }
-  ;
-
-
-  $err = eval {
-    require Class::XSAccessor;
-    Class::XSAccessor->VERSION($__minimum_xsa_version);
-    require Sub::Name;
-    1;
-  } ? undef : do {
-    delete $INC{'Sub/Name.pm'};   # because older perls suck
-    delete $INC{'Class/XSAccessor.pm'};
-    $@;
-  };
-  *__CAG_ENV__::NO_CXSA = $err
-    ? sub () { $err }
-    : sub () { 0 }
-  ;
-
-
-  *__CAG_ENV__::BROKEN_GOTO = ($] < '5.008009')
-    ? sub () { 1 }
-    : sub () { 0 }
-  ;
-
-
-  *__CAG_ENV__::UNSTABLE_DOLLARAT = ($] < '5.013002')
-    ? sub () { 1 }
-    : sub () { 0 }
-  ;
-
-
-  *__CAG_ENV__::TRACK_UNDEFER_FAIL = (
-    $INC{'Test/Builder.pm'} || $INC{'Test/Builder2.pm'}
-      and
-    $0 =~ m|^ x?t / .+ \.t $|x
-  ) ? sub () { 1 }
-    : sub () { 0 }
-  ;
-}
-
 # Autodetect unless flag supplied
 my $xsa_autodetected;
 if (! defined $USE_XS) {
@@ -589,58 +585,71 @@ if (! defined $USE_XS) {
   $xsa_autodetected++;
 }
 
+my $perlstring;
+if ($] < '5.008') {
+  require Data::Dumper;
+  my $d = Data::Dumper->new([])->Indent(0)->Purity(0)->Pad('')->Useqq(1)->Terse(1)->Freezer('')->Toaster('');
+  $perlstring = sub { $d->Values([shift])->Dump };
+}
+else {
+  require B;
+  $perlstring = \&B::perlstring;
+}
+
+
 my $maker_templates = {
   rw => {
-    xs_call => 'accessors',
-    pp_code => sub {
-      my $set = "set_$_[0]";
-      my $get = "get_$_[0]";
-      my $field = $_[1];
-      $field =~ s/'/\\'/g;
+    cxsa_call => 'accessors',
+    pp_generator => sub {
+      # my ($group, $fieldname) = @_;
+      my $quoted_fieldname = $perlstring->($_[1]);
+      sprintf <<'EOS', ($_[0], $quoted_fieldname) x 2;
+
+@_ > 1
+  ? shift->set_%s(%s, @_)
+  : shift->get_%s(%s)
+EOS
 
-      "
-        \@_ != 1
-          ? shift->$set('$field', \@_)
-          : shift->$get('$field')
-      "
     },
   },
   ro => {
-    xs_call => 'getters',
-    pp_code => sub {
-      my $get = "get_$_[0]";
-      my $field = $_[1];
-      $field =~ s/'/\\'/g;
+    cxsa_call => 'getters',
+    pp_generator => sub {
+      # my ($group, $fieldname) = @_;
+      my $quoted_fieldname = $perlstring->($_[1]);
+      sprintf  <<'EOS', $_[0], $quoted_fieldname;
+
+@_ > 1
+  ? do {
+    my ($meth) = (caller(0))[3] =~ /([^\:]+)$/;
+    my $class = length( ref($_[0]) ) ? ref($_[0]) : $_[0];
+    Carp::croak(
+      "'$meth' cannot alter its value (read-only attribute of class $class)"
+    );
+  }
+  : shift->get_%s(%s)
+EOS
 
-      "
-        \@_ == 1
-          ? shift->$get('$field')
-          : do {
-            my \$caller = caller;
-            my \$class = ref \$_[0] || \$_[0];
-            Carp::croak(\"'\$caller' cannot alter the value of '$field' \".
-                        \"(read-only attributes of class '\$class')\");
-          }
-      "
     },
   },
   wo => {
-    xs_call => 'setters',
-    pp_code => sub {
-      my $set = "set_$_[0]";
-      my $field = $_[1];
-      $field =~ s/'/\\'/g;
+    cxsa_call => 'setters',
+    pp_generator => sub {
+      # my ($group, $fieldname) = @_;
+      my $quoted_fieldname = $perlstring->($_[1]);
+      sprintf  <<'EOS', $_[0], $quoted_fieldname;
+
+@_ > 1
+  ? shift->set_%s(%s, @_)
+  : do {
+    my ($meth) = (caller(0))[3] =~ /([^\:]+)$/;
+    my $class = length( ref($_[0]) ) ? ref($_[0]) : $_[0];
+    Carp::croak(
+      "'$meth' cannot access its value (write-only attribute of class $class)"
+    );
+  }
+EOS
 
-      "
-        \@_ != 1
-          ? shift->$set('$field', \@_)
-          : do {
-            my \$caller = caller;
-            my \$class = ref \$_[0] || \$_[0];
-            Carp::croak(\"'\$caller' cannot access the value of '$field' \".
-                        \"(write-only attributes of class '\$class')\");
-          }
-      "
     },
   },
 };
@@ -656,9 +665,7 @@ my $original_simple_setter = __PACKAGE__->can ('set_simple');
 # Note!!! Unusual signature
 $gen_accessor = sub {
   my ($type, $class, $group, $field, $methname) = @_;
-  if (my $c = Scalar::Util::blessed( $class )) {
-    $class = $c;
-  }
+  $class = ref $class if length ref $class;
 
   # When installing an XSA simple accessor, we need to make sure we are not
   # short-circuiting a (compile or runtime) get_simple/set_simple override.
@@ -675,7 +682,7 @@ $gen_accessor = sub {
 
     my ($expected_cref, $cached_implementation);
     my $ret = $expected_cref = sub {
-      my $current_class = Scalar::Util::blessed( $_[0] ) || $_[0];
+      my $current_class = length (ref ($_[0] ) ) ? ref ($_[0]) : $_[0];
 
       # $cached_implementation will be set only if the shim got
       # 'around'ed, in which case it is handy to avoid re-running
@@ -699,7 +706,7 @@ $gen_accessor = sub {
           Class::XSAccessor->import(
             replace => 1,
             class => '__CAG__XSA__BREEDER__',
-            $maker_templates->{$type}{xs_call} => {
+            $maker_templates->{$type}{cxsa_call} => {
               $methname => $field,
             },
           );
@@ -750,7 +757,7 @@ $gen_accessor = sub {
             "Deferred version of method $cframe[3] invoked more than once (originally "
           . "invoked at $already_seen). This is a strong indication your code has "
           . 'cached the original ->can derived method coderef, and is using it instead '
-          . 'of the proper method re-lookup, causing performance regressions'
+          . 'of the proper method re-lookup, causing minor performance regressions'
           );
         }
         else {
@@ -788,7 +795,7 @@ $gen_accessor = sub {
   # no Sub::Name - just install the coderefs directly (compiling every time)
   elsif (__CAG_ENV__::NO_SUBNAME) {
     my $src = $accessor_maker_cache->{source}{$type}{$group}{$field} ||=
-      $maker_templates->{$type}{pp_code}->($group, $field);
+      $maker_templates->{$type}{pp_generator}->($group, $field);
 
     no warnings 'redefine';
     local $@ if __CAG_ENV__::UNSTABLE_DOLLARAT;
@@ -801,7 +808,7 @@ $gen_accessor = sub {
   else {
     ($accessor_maker_cache->{pp}{$type}{$group}{$field} ||= do {
       my $src = $accessor_maker_cache->{source}{$type}{$group}{$field} ||=
-        $maker_templates->{$type}{pp_code}->($group, $field);
+        $maker_templates->{$type}{pp_generator}->($group, $field);
 
       local $@ if __CAG_ENV__::UNSTABLE_DOLLARAT;
       eval "sub { my \$dummy; sub { \$dummy if 0; $src } }" or die $@;
